/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 * Copyright © 2007 Gerd Kohlberger <lowfi@chello.at>
 */

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "mt-main.h"
#include "mt-dbus.h"

extern GladeXML *xml;

const gchar *img_widgets[] = {
    "single_click_img",
    "double_click_img",
    "drag_click_img",
    "right_click_img"
};

void
mt_ctw_update_sensitivity (MTClosure *mt)
{
    GtkWidget *w;
    gboolean sensitive = FALSE;

    if (mt->dwell_enabled && mt->dwell_mode == DWELL_MODE_CTW)
	sensitive = TRUE;

    w = glade_xml_get_widget (xml, "button_box");
    gtk_widget_set_sensitive (w, sensitive);
}

static void
ctw_button_cb (GtkToggleButton *button, gpointer data)
{
    MTClosure *mt = (MTClosure *) data;

    if (gtk_toggle_button_get_active (button)) {
	GSList *group;

	group = gtk_radio_button_get_group (GTK_RADIO_BUTTON(button));
	mt->dwell_cct = g_slist_index (group, (gconstpointer) button);

	mt_dbus_send_signal (mt->conn, CLICK_TYPE_SIGNAL, mt->dwell_cct);
    }
}

static gboolean
ctw_context_menu (GtkWidget *widget, GdkEventButton *bev, gpointer data)
{
    if (bev->button == 3) {
	gtk_menu_popup (GTK_MENU(glade_xml_get_widget (xml, "popup")),
			0, 0, 0, 0, bev->button, bev->time);

	return TRUE;
    }

    return FALSE;
}

static void
change_button_style (ButtonStyle style)
{
    GtkWidget *icon, *label;
    gint i;
    const gchar *labels[] = {
	"single_click_label",
	"double_click_label",
	"drag_click_label",
	"right_click_label"
    };

    for (i = 0; i < N_CLICK_TYPES; i++) {
	label = glade_xml_get_widget (xml, labels[i]);
	icon = glade_xml_get_widget (xml, img_widgets[i]);

	switch (style) {
	case BUTTON_STYLE_BOTH:
	    g_object_set (icon, "yalign", 1.0, NULL);
	    gtk_widget_show (icon);
	    g_object_set (label, "yalign", 0.0, NULL);
	    gtk_widget_show (label);
	    break;
	case BUTTON_STYLE_TEXT:
	    gtk_widget_hide (icon);
	    g_object_set (icon, "yalign", 0.5, NULL);
	    gtk_widget_show (label);
	    g_object_set (label, "yalign", 0.5, NULL);
	    break;
	case BUTTON_STYLE_ICON:
	    gtk_widget_show (icon);
	    g_object_set (icon, "yalign", 0.5, NULL);
	    gtk_widget_hide (label);
	    g_object_set (label, "yalign", 0.5, NULL);
	default:
	    break;
	}
    }
}

static void
ctw_menu_toggled (GtkCheckMenuItem *item, gpointer data)
{
    MTClosure *mt = (MTClosure *) data;
    GSList *group;

    if (!gtk_check_menu_item_get_active (item))
	return;

    group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM(item));
    mt->style = g_slist_index (group, (gconstpointer) item);
    gconf_client_set_int (mt->client,
			  MT_GCONF_HOME "/button_layout",
			  (gint) mt->style,
			  NULL);

    change_button_style (mt->style);
}

static gboolean
ctw_delete_cb (GtkWidget *win, GdkEvent *ev, gpointer data)
{
    GtkWidget *w;

    w = glade_xml_get_widget (xml, "dwell_show_ctw");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(w), FALSE);

    return TRUE;
}

void
mt_ctw_init (MTClosure *mt)
{
    GtkWidget *w;
    gint i;
    gchar *path;
    GSList *group;
    gpointer data;
    const gchar *buttons[] = {
	"single_click",
	"double_click",
	"drag_click",
	"right_click"
    };
    const gchar *pngs[] = {
	"human-single.png",
	"human-double.png",
	"human-drag.png",
	"human-right.png"
    };

    w = glade_xml_get_widget (xml, "click-type");
    gtk_window_stick (GTK_WINDOW(w));
    gtk_window_set_keep_above (GTK_WINDOW(w), TRUE);
    g_signal_connect (G_OBJECT(w), "delete-event", 
		      G_CALLBACK(ctw_delete_cb), NULL);

    if (mt->dwell_show_ctw)
	gtk_widget_show (w);

    for (i = 0; i < N_CLICK_TYPES; i++) {
	w = glade_xml_get_widget (xml, buttons[i]);
	gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON(w), FALSE);

	g_signal_connect (G_OBJECT(w), "toggled", 
			  G_CALLBACK(ctw_button_cb), (gpointer) mt);
	g_signal_connect (G_OBJECT(w), "button-press-event", 
			  G_CALLBACK(ctw_context_menu), NULL);

	/* icons */
	w = glade_xml_get_widget (xml, img_widgets[i]);
	path = g_strdup_printf ("%s/%s", PIXMAPS_PATH, pngs[i]);
	gtk_image_set_from_file (GTK_IMAGE(w), path);
	g_free (path);
    }

    w = glade_xml_get_widget (xml, "both_item");

    group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM(w));
    data = g_slist_nth_data (group, mt->style);
    gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM(data), TRUE);

    g_signal_connect (G_OBJECT(w), "toggled", 
		      G_CALLBACK(ctw_menu_toggled), (gpointer) mt);
    w = glade_xml_get_widget (xml, "text_item");
    g_signal_connect (G_OBJECT(w), "toggled", 
		      G_CALLBACK(ctw_menu_toggled), (gpointer) mt);
    w = glade_xml_get_widget (xml, "icon_item");
    g_signal_connect (G_OBJECT(w), "toggled", 
		      G_CALLBACK(ctw_menu_toggled), (gpointer) mt);

    change_button_style (mt->style);
    mt_ctw_update_sensitivity (mt);
}
