/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 * Copyright © 2007 Gerd Kohlberger <lowfi@chello.at>
 */

#include <gtk/gtk.h>
#include <glade/glade.h>

#include <cspi/spi.h>

#include <libgnome/gnome-program.h>
#include <libgnome/gnome-help.h>
#include <libgnomeui/gnome-ui-init.h>

#include <math.h>

#include "mt-main.h"
#include "mt-session.h"
#include "mt-dbus.h"
#include "mt-trayicon.h"
#include "mt-preferences.h"
#include "mt-ct-window.h"

GladeXML *xml = NULL;

static AccessibleDeviceListener *button_listener;
static AccessibleEventListener  *motion_listener;

static void start_dwell_timer (MTClosure *mt);
static void stop_timer	      (GTimer *timer, guint *tid);

static void
mt_set_grab_cursor (gboolean set)
{
    GdkCursor *cursor;
    GdkWindow *root;

    root = gdk_get_default_root_window ();

    if (set)
	cursor = gdk_cursor_new (GDK_FLEUR);
    else
	cursor = gdk_cursor_new (GDK_ARROW);

    gdk_window_set_cursor (root, cursor);
    gdk_cursor_unref (cursor);
}

static void
dwell_restore_single_click (MTClosure *mt)
{
    if (mt->dwell_mode == DWELL_MODE_CTW) {
	GtkWidget *button;

	button = glade_xml_get_widget (xml, "single_click");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), TRUE);

	mt_dbus_send_signal (mt->conn, RESTORE_SIGNAL, 0);
    }
}

static void
dwell_do_pointer_click (MTClosure *mt, gint x, gint y)
{
    switch (mt->dwell_cct) {
    case DWELL_CLICK_TYPE_SINGLE:
	SPI_generateMouseEvent (x, y, "b1c");
	break;
    case DWELL_CLICK_TYPE_DOUBLE:
	SPI_generateMouseEvent (x, y, "b1d");
	dwell_restore_single_click (mt);
	break;
    case DWELL_CLICK_TYPE_DRAG:
	SPI_generateMouseEvent (x, y, "b1p");
	mt->dwell_drag_started = TRUE;

	mt_set_grab_cursor (TRUE);
	break;
    case DWELL_CLICK_TYPE_RIGHT:
	SPI_generateMouseEvent (x, y, "b3c");
	dwell_restore_single_click (mt);
    default:
	break;
    }
}

static inline gboolean
within_tolerance (MTClosure *mt, gint x, gint y)
{
    gdouble distance;

    distance = sqrt(pow(x - mt->pointer_x, 2.) + pow(y - mt->pointer_y, 2.));

    if (distance < mt->threshold)
	return TRUE;

    return FALSE;
}

static gboolean
analyze_direction (MTClosure *mt)
{
    GestureDirection gd;
    gint i, x, y, dx, dy;

    gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, NULL);

    dx = ABS(x - mt->pointer_x);
    dy = ABS(y - mt->pointer_y);

    if (within_tolerance (mt, x, y))
	return FALSE;

    /* find direction */
    if (x < mt->pointer_x)
	if (y < mt->pointer_y)
	    if (dx < dy)
		gd = DIRECTION_UP;
	    else
		gd = DIRECTION_LEFT;
	else
	    if (dx < dy)
		gd = DIRECTION_DOWN;
	    else
		gd = DIRECTION_LEFT;
    else
	if (y < mt->pointer_y)
	    if (dx < dy)
		gd = DIRECTION_UP;
	    else
		gd = DIRECTION_RIGHT;
	else
	    if (dx < dy)
		gd = DIRECTION_DOWN;
	    else
		gd = DIRECTION_RIGHT;

    /* get click type for direction */
    for (i = 0; i < N_CLICK_TYPES; i++)
	if (mt->dwell_dirs[i] == gd) {
	    mt->dwell_cct = i;
	    return TRUE;
	}

    return FALSE;
}

static void
dwell_start_gesture (MTClosure *mt)
{
    GdkCursor *cursor;
    GdkWindow *root;
    GdkGrabStatus stat;

    cursor = gdk_cursor_new (GDK_CROSS);
    root = gdk_get_default_root_window ();
    stat = gdk_pointer_grab (root,
			     FALSE, 
			     GDK_POINTER_MOTION_MASK, 
			     NULL,
			     cursor,
			     GDK_CURRENT_TIME);
    gdk_cursor_unref (cursor);

    if (stat != GDK_GRAB_SUCCESS)
	return;

    mt->dwell_gesture_started = TRUE;
    start_dwell_timer (mt);
}

static void
dwell_stop_gesture (MTClosure *mt)
{
    stop_timer (mt->dwell_timer, &mt->dwell_tid);
    mt->dwell_gesture_started = FALSE;

    gdk_pointer_ungrab (GDK_CURRENT_TIME);

    if (analyze_direction (mt)) {
	/* gesture: context menu workaround */
	if (mt->dwell_cct == DWELL_CLICK_TYPE_RIGHT)
	    mt->context_menu_opened = TRUE;

	dwell_do_pointer_click (mt, mt->pointer_x, mt->pointer_y);

	if (mt->context_menu_opened) {
	    mt->dwell_mode = DWELL_MODE_CTW;
	    mt->dwell_cct = DWELL_CLICK_TYPE_SINGLE;
	}
    }
}

/* timeout functions */
static gboolean
dwell_check_time (gpointer data)
{
    MTClosure *mt = (MTClosure *) data;

    if (g_timer_elapsed (mt->dwell_timer, NULL) > mt->dwell_time) {
	gint x, y;

	gdk_display_get_pointer (gdk_display_get_default(), NULL, &x, &y, NULL);

	/* stop active drag */
	if (mt->dwell_drag_started) {
	    SPI_generateMouseEvent (x, y, "b1r");
	    mt->dwell_drag_started = FALSE;

	    mt_set_grab_cursor (FALSE);

	    dwell_restore_single_click (mt);
	    stop_timer (mt->dwell_timer, &mt->dwell_tid);

	    return FALSE;
	}

	/* click type window mode */
	if (mt->dwell_mode == DWELL_MODE_CTW) {
	    dwell_do_pointer_click (mt, x, y);
	    stop_timer (mt->dwell_timer, &mt->dwell_tid);

	    /* gesture: context menu workaround */
	    if (mt->context_menu_opened) {
		mt->context_menu_opened = FALSE;
		mt->dwell_mode = DWELL_MODE_GESTURE;
	    }

	    return FALSE;
	}

	/* gesture mode */
	if (mt->dwell_mode == DWELL_MODE_GESTURE) {
	    if (mt->dwell_gesture_started) {
		dwell_stop_gesture (mt);
		return FALSE;
	    }
	    else if (!mt->dwell_enabled) {
		stop_timer (mt->dwell_timer, &mt->dwell_tid);
		return FALSE;
	    }
	    else
		dwell_start_gesture (mt);
	}
    }

    return TRUE;
}

static gboolean
delay_check_time (gpointer data)
{
    MTClosure *mt = (MTClosure *) data;

    if (g_timer_elapsed (mt->delay_timer, NULL) > mt->delay_time) {
	SPI_generateMouseEvent (0, 0, "b1r");
	SPI_generateMouseEvent (mt->pointer_x, mt->pointer_y, "b3c");

	return FALSE;
    }

    return TRUE;
}

/* timer stuff */
static void
start_dwell_timer (MTClosure *mt)
{
    g_timer_start (mt->dwell_timer);

    if (!mt->dwell_tid)
	mt->dwell_tid = g_timeout_add (100,
				       (GSourceFunc) dwell_check_time,
				       (gpointer) mt);
}

static void
start_delay_timer (MTClosure *mt)
{
    g_timer_start (mt->delay_timer);

    if (!mt->delay_tid)
	mt->delay_tid = g_timeout_add (100,
				       (GSourceFunc) delay_check_time,
				       (gpointer) mt);
}

static void
stop_timer (GTimer *timer, guint *tid)
{
    g_timer_stop (timer);
    g_timer_reset (timer);

    if (*tid) {
	g_source_remove (*tid);
	*tid = 0;
    }
}

/* at-spi callbacks */
static void
spi_motion_event (const AccessibleEvent *event, void *data)
{
    MTClosure *mt = (MTClosure *) data;

    if (mt->dwell_enabled)
	if (!within_tolerance (mt, event->detail1, event->detail2))
	    if (!mt->dwell_gesture_started) {
		mt->pointer_x = (gint) event->detail1;
		mt->pointer_y = (gint) event->detail2;

		start_dwell_timer (mt);
	    }

    if (mt->delay_tid)
	if (!within_tolerance (mt, event->detail1, event->detail2))
	    stop_timer (mt->delay_timer, &mt->delay_tid);
}

static SPIBoolean
spi_button_event (const AccessibleDeviceEvent *event, void *data)
{
    MTClosure *mt = (MTClosure *) data;

    if (mt->delay_enabled && event->keycode == 1) {
	switch (event->type) {
	case SPI_BUTTON_PRESSED:
	    gdk_display_get_pointer (gdk_display_get_default (),
				     NULL,
				     &mt->pointer_x,
				     &mt->pointer_y,
				     NULL);
	    start_delay_timer (mt);
	    return TRUE;
	case SPI_BUTTON_RELEASED:
	    if (mt->delay_tid)
		stop_timer (mt->delay_timer, &mt->delay_tid);

	    return TRUE;
	default:
	    break;
	}
    }

    return FALSE;
}

gint
mt_show_dialog (const gchar *primary,
		const gchar *secondary,
		GtkMessageType type)
{
    GtkWidget *dialog;
    gint ret;

    dialog = gtk_message_dialog_new (NULL,
				     GTK_DIALOG_MODAL,
				     type,
				     GTK_BUTTONS_NONE,
				     primary);
    gtk_window_set_title (GTK_WINDOW(dialog), _("MouseTweaks"));
    gtk_window_set_icon_name (GTK_WINDOW(dialog), MT_ICON_NAME);
    gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG(dialog),
					      secondary);

    if (type == GTK_MESSAGE_QUESTION)
	gtk_dialog_add_buttons (GTK_DIALOG(dialog),
				GTK_STOCK_YES, GTK_RESPONSE_YES,
				GTK_STOCK_NO, GTK_RESPONSE_NO,
				NULL);
    else
	gtk_dialog_add_button (GTK_DIALOG(dialog),
			       GTK_STOCK_OK, GTK_RESPONSE_OK);

    gtk_widget_show_all (dialog);
    ret = gtk_dialog_run (GTK_DIALOG(dialog));
    gtk_widget_destroy (dialog);

    return ret;
}

void
mt_show_preferences (void)
{
    GtkWidget *w;

    w = glade_xml_get_widget (xml, "preferences");

    if (GTK_WIDGET_VISIBLE(w))
	gtk_window_present (GTK_WINDOW(w));
    else
	gtk_widget_show (w);
}

void
spi_shutdown (void)
{
    SPI_deregisterGlobalEventListener (motion_listener, "mouse:abs");
    AccessibleEventListener_unref (motion_listener);

    SPI_deregisterDeviceEventListener (button_listener, NULL);
    AccessibleDeviceListener_unref (button_listener);

    SPI_event_quit ();
}

static void
mt_gconf_read_options (MTClosure *mt)
{
    mt->threshold = gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/threshold", NULL);
    mt->delay_enabled = gconf_client_get_bool
	(mt->client, MT_GCONF_HOME "/delay_enabled", NULL);
    mt->delay_time = gconf_client_get_float
	(mt->client, MT_GCONF_HOME "/delay_time", NULL);
    mt->dwell_enabled = gconf_client_get_bool
	(mt->client, MT_GCONF_HOME "/dwell_enabled", NULL);
    mt->dwell_time = gconf_client_get_float
	(mt->client, MT_GCONF_HOME "/dwell_time", NULL);
    mt->dwell_show_ctw = gconf_client_get_bool
	(mt->client, MT_GCONF_HOME "/dwell_show_ctw", NULL);
    mt->dwell_mode = gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/dwell_mode", NULL);
    mt->dwell_dirs[DWELL_CLICK_TYPE_SINGLE] = gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/dwell_gesture_single", NULL);
    mt->dwell_dirs[DWELL_CLICK_TYPE_DOUBLE] = gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/dwell_gesture_double", NULL);
    mt->dwell_dirs[DWELL_CLICK_TYPE_DRAG] = gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/dwell_gesture_drag", NULL);
    mt->dwell_dirs[DWELL_CLICK_TYPE_RIGHT] = gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/dwell_gesture_right", NULL);
    mt->style = (ButtonStyle) gconf_client_get_int
	(mt->client, MT_GCONF_HOME "/button_layout", NULL);
}

static MTClosure *
mt_closure_init (void)
{
    MTClosure *mt;

    mt = g_new0 (MTClosure, 1);
    if (!mt)
	return NULL;

    mt->delay_timer = g_timer_new ();
    mt->dwell_timer = g_timer_new ();

    stop_timer (mt->delay_timer, &mt->delay_tid);
    stop_timer (mt->dwell_timer, &mt->dwell_tid);

    mt->client = gconf_client_get_default ();
    mt->dwell_cct = DWELL_CLICK_TYPE_SINGLE;

    return mt;
}

static void
mt_closure_free (MTClosure *mt)
{
    g_timer_destroy (mt->dwell_timer);
    g_timer_destroy (mt->delay_timer);

    g_object_unref (mt->client);

    g_free (mt);
}

int
main (int argc, char **argv)
{
    GnomeProgram *program;
    MTClosure *mt;
    gboolean dwell = FALSE;
    gboolean delay = FALSE;
    gboolean ctw = FALSE;
    gchar *mode = NULL;
    GOptionContext *context;
    GOptionEntry entries[] = {
	{"enable-dwell", 0, 0, G_OPTION_ARG_NONE, &dwell,
	    _("Enable Dwell Click"), 0},
	{"enable-delay", 0, 0, G_OPTION_ARG_NONE, &delay,
	    _("Enable Delay Click"), 0},
	{"dwell-mode", 0, 0, G_OPTION_ARG_STRING, &mode,
	    _("Use dwell mode"), "(window|gesture)"},
	{"show-ctw", 0, 0, G_OPTION_ARG_NONE, &ctw,
	    "Show click type window", 0},
	{NULL}
    };

    gtk_init (&argc, &argv);

    bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
    textdomain (GETTEXT_PACKAGE);

    context = g_option_context_new ("");
    g_option_context_add_main_entries (context, entries, GETTEXT_PACKAGE);
    g_option_context_add_group (context, gtk_get_option_group (FALSE));
    g_option_context_parse (context, &argc, &argv, NULL);
    g_option_context_free (context);

    program = gnome_program_init (PACKAGE, VERSION,
				  LIBGNOMEUI_MODULE,
				  argc, argv,
				  GNOME_PARAM_APP_DATADIR, MT_DATADIR,
				  GNOME_PARAM_NONE);

    mt_session_init (argc, argv);

    if (SPI_init ()) {
	mt_show_dialog (_("Assistive Technologies not enabled"),
			_("MouseTweaks requires Assistive Technologies."),
			GTK_MESSAGE_ERROR);
	g_object_unref (program);
	return 1;
    }

    mt = mt_closure_init ();
    if (!mt) {
	mt_show_dialog (_("Internal Error"),
			_("Couldn't allocate memory."),
			GTK_MESSAGE_ERROR);
	g_object_unref (program);
	SPI_exit ();
	return 1;
    }

    xml = glade_xml_new (GLADE_PATH "/mousetweaks.glade", NULL, NULL);
    if (!xml) {
	mt_show_dialog (_("Internal Error"),
			_("Failed to load glade interface file."),
			GTK_MESSAGE_ERROR);
	goto CLEANUP;
    }

    mt->conn = mt_dbus_init (mt);
    if (!mt->conn)
	goto CLEANUP;

    mt_dbus_send_signal (mt->conn, ACTIVE_SIGNAL, 1);

    /* add at-spi listeners */
    motion_listener = SPI_createAccessibleEventListener (spi_motion_event,
							 (void *) mt);
    SPI_registerGlobalEventListener (motion_listener, "mouse:abs");
    button_listener = SPI_createAccessibleDeviceListener (spi_button_event,
							  (void *) mt);
    SPI_registerDeviceEventListener (button_listener,
				     SPI_BUTTON_PRESSED | SPI_BUTTON_RELEASED,
				     NULL);

    mt_gconf_read_options (mt);

    /* override options from command-line */
    if (dwell)
	mt->dwell_enabled = TRUE;
    if (delay)
	mt->delay_enabled = TRUE;
    if (ctw)
	mt->dwell_show_ctw = TRUE;
    if (mode) {
	if (g_str_equal (mode, "gesture"))
	    mt->dwell_mode = 0;
	if (g_str_equal (mode, "window"))
	    mt->dwell_mode = 1;
    }

    mt_trayicon_init (mt);
    mt_preferences_init (mt);
    mt_ctw_init (mt);

    SPI_event_main ();

    mt_dbus_send_signal (mt->conn, ACTIVE_SIGNAL, 0);
    mt_dbus_fini (mt->conn);

CLEANUP:
    SPI_exit ();
    mt_closure_free (mt);
    g_object_unref (program);

    return 0;
}
